// Generated by CoffeeScript 2.2.1
(function() {
  var XMLAttribute, XMLElement, XMLNode, isFunction, isObject,
    hasProp = {}.hasOwnProperty;

  ({isObject, isFunction} = require('./Utility'));

  XMLNode = require('./XMLNode');

  XMLAttribute = require('./XMLAttribute');

  // Represents an element of the XML document
  module.exports = XMLElement = class XMLElement extends XMLNode {
    // Initializes a new instance of `XMLElement`

    // `parent` the parent node
    // `name` element name
    // `attributes` an object containing name/value pairs of attributes
    constructor(parent, name, attributes) {
      super(parent);
      if (name == null) {
        throw new Error("Missing element name");
      }
      this.name = this.stringify.eleName(name);
      this.attributes = {};
      if (attributes != null) {
        this.attribute(attributes);
      }
      // set properties if this is the root node
      if (parent.isDocument) {
        this.isRoot = true;
        this.documentObject = parent;
        parent.rootObject = this;
      }
    }

    // Creates and returns a deep clone of `this`

    clone() {
      var att, attName, clonedSelf, ref;
      clonedSelf = Object.create(this);
      // remove document element
      if (clonedSelf.isRoot) {
        clonedSelf.documentObject = null;
      }
      // clone attributes
      clonedSelf.attributes = {};
      ref = this.attributes;
      for (attName in ref) {
        if (!hasProp.call(ref, attName)) continue;
        att = ref[attName];
        clonedSelf.attributes[attName] = att.clone();
      }
      // clone child nodes
      clonedSelf.children = [];
      this.children.forEach(function(child) {
        var clonedChild;
        clonedChild = child.clone();
        clonedChild.parent = clonedSelf;
        return clonedSelf.children.push(clonedChild);
      });
      return clonedSelf;
    }

    // Adds or modifies an attribute

    // `name` attribute name
    // `value` attribute value
    attribute(name, value) {
      var attName, attValue;
      if (name != null) {
        name = name.valueOf();
      }
      if (isObject(name)) { // expand if object
        for (attName in name) {
          if (!hasProp.call(name, attName)) continue;
          attValue = name[attName];
          this.attribute(attName, attValue);
        }
      } else {
        if (isFunction(value)) {
          value = value.apply();
        }
        if (!this.options.skipNullAttributes || (value != null)) {
          this.attributes[name] = new XMLAttribute(this, name, value);
        }
      }
      return this;
    }

    // Removes an attribute

    // `name` attribute name
    removeAttribute(name) {
      var attName, i, len;
      if (name == null) {
        throw new Error("Missing attribute name");
      }
      name = name.valueOf();
      if (Array.isArray(name)) { // expand if array
        for (i = 0, len = name.length; i < len; i++) {
          attName = name[i];
          delete this.attributes[attName];
        }
      } else {
        delete this.attributes[name];
      }
      return this;
    }

    // Converts the XML fragment to string

    // `options.pretty` pretty prints the result
    // `options.indent` indentation for pretty print
    // `options.offset` how many indentations to add to every line for pretty print
    // `options.newline` newline sequence for pretty print
    // `options.allowEmpty` do not self close empty element tags
    toString(options) {
      return this.options.writer.set(options).element(this);
    }

    // Aliases
    att(name, value) {
      return this.attribute(name, value);
    }

    a(name, value) {
      return this.attribute(name, value);
    }

  };

}).call(this);
