// Generated by CoffeeScript 2.2.1
(function() {
  var XMLCData, XMLComment, XMLDeclaration, XMLDocType, XMLElement, XMLNode, XMLProcessingInstruction, XMLRaw, XMLText, isEmpty, isFunction, isObject,
    hasProp = {}.hasOwnProperty,
    splice = [].splice;

  ({isObject, isFunction, isEmpty} = require('./Utility'));

  XMLElement = null;

  XMLCData = null;

  XMLComment = null;

  XMLDeclaration = null;

  XMLDocType = null;

  XMLRaw = null;

  XMLText = null;

  XMLProcessingInstruction = null;

  // Represents a generic XMl element
  module.exports = XMLNode = class XMLNode {
    // Initializes a new instance of `XMLNode`

    // `parent` the parent node
    constructor(parent) {
      this.parent = parent;
      if (this.parent) {
        this.options = this.parent.options;
        this.stringify = this.parent.stringify;
      }
      this.children = [];
      // first execution, load dependencies that are otherwise
      // circular (so we can't load them at the top)
      if (!XMLElement) {
        XMLElement = require('./XMLElement');
        XMLCData = require('./XMLCData');
        XMLComment = require('./XMLComment');
        XMLDeclaration = require('./XMLDeclaration');
        XMLDocType = require('./XMLDocType');
        XMLRaw = require('./XMLRaw');
        XMLText = require('./XMLText');
        XMLProcessingInstruction = require('./XMLProcessingInstruction');
      }
    }

    // Creates a child element node

    // `name` node name or an object describing the XML tree
    // `attributes` an object containing name/value pairs of attributes
    // `text` element text
    element(name, attributes, text) {
      var childNode, item, j, k, key, lastChild, len, len1, val;
      lastChild = null;
      if (attributes == null) {
        attributes = {};
      }
      attributes = attributes.valueOf();
      // swap argument order: text <-> attributes
      if (!isObject(attributes)) {
        [text, attributes] = [attributes, text];
      }
      if (name != null) {
        name = name.valueOf();
      }
      // expand if array
      if (Array.isArray(name)) {
        for (j = 0, len = name.length; j < len; j++) {
          item = name[j];
          lastChild = this.element(item);
        }
      // evaluate if function
      } else if (isFunction(name)) {
        lastChild = this.element(name.apply());
      // expand if object
      } else if (isObject(name)) {
        for (key in name) {
          if (!hasProp.call(name, key)) continue;
          val = name[key];
          if (isFunction(val)) {
            // evaluate if function
            val = val.apply();
          }
          if ((isObject(val)) && (isEmpty(val))) {
            // skip empty or null objects and arrays
            val = null;
          }
          // assign attributes
          if (!this.options.ignoreDecorators && this.stringify.convertAttKey && key.indexOf(this.stringify.convertAttKey) === 0) {
            lastChild = this.attribute(key.substr(this.stringify.convertAttKey.length), val);
          // expand list by creating child nodes
          } else if (!this.options.separateArrayItems && Array.isArray(val)) {
            for (k = 0, len1 = val.length; k < len1; k++) {
              item = val[k];
              childNode = {};
              childNode[key] = item;
              lastChild = this.element(childNode);
            }
          // expand child nodes under parent
          } else if (isObject(val)) {
            lastChild = this.element(key);
            lastChild.element(val);
          } else {
            // text node
            lastChild = this.element(key, val);
          }
        }
      } else {
        // text node
        if (!this.options.ignoreDecorators && this.stringify.convertTextKey && name.indexOf(this.stringify.convertTextKey) === 0) {
          lastChild = this.text(text);
        // cdata node
        } else if (!this.options.ignoreDecorators && this.stringify.convertCDataKey && name.indexOf(this.stringify.convertCDataKey) === 0) {
          lastChild = this.cdata(text);
        // comment node
        } else if (!this.options.ignoreDecorators && this.stringify.convertCommentKey && name.indexOf(this.stringify.convertCommentKey) === 0) {
          lastChild = this.comment(text);
        // raw text node
        } else if (!this.options.ignoreDecorators && this.stringify.convertRawKey && name.indexOf(this.stringify.convertRawKey) === 0) {
          lastChild = this.raw(text);
        // processing instruction
        } else if (!this.options.ignoreDecorators && this.stringify.convertPIKey && name.indexOf(this.stringify.convertPIKey) === 0) {
          lastChild = this.instruction(name.substr(this.stringify.convertPIKey.length), text);
        } else {
          // element node
          lastChild = this.node(name, attributes, text);
        }
      }
      if (lastChild == null) {
        throw new Error("Could not create any elements with: " + name);
      }
      return lastChild;
    }

    // Creates a child element node before the current node

    // `name` node name or an object describing the XML tree
    // `attributes` an object containing name/value pairs of attributes
    // `text` element text
    insertBefore(name, attributes, text) {
      var child, i, removed;
      if (this.isRoot) {
        throw new Error("Cannot insert elements at root level");
      }
      // temporarily remove children starting *with* this
      i = this.parent.children.indexOf(this);
      removed = this.parent.children.splice(i);
      // add the new child
      child = this.parent.element(name, attributes, text);
      // add back removed children after new child
      Array.prototype.push.apply(this.parent.children, removed);
      return child;
    }

    // Creates a child element node after the current node

    // `name` node name or an object describing the XML tree
    // `attributes` an object containing name/value pairs of attributes
    // `text` element text
    insertAfter(name, attributes, text) {
      var child, i, removed;
      if (this.isRoot) {
        throw new Error("Cannot insert elements at root level");
      }
      // temporarily remove children starting *after* this
      i = this.parent.children.indexOf(this);
      removed = this.parent.children.splice(i + 1);
      // add the new child
      child = this.parent.element(name, attributes, text);
      // add back removed children after new child
      Array.prototype.push.apply(this.parent.children, removed);
      return child;
    }

    // Deletes a child element node

    remove() {
      var i, ref;
      if (this.isRoot) {
        throw new Error("Cannot remove the root element");
      }
      i = this.parent.children.indexOf(this);
      splice.apply(this.parent.children, [i, i - i + 1].concat(ref = [])), ref;
      return this.parent;
    }

    // Creates a node

    // `name` name of the node
    // `attributes` an object containing name/value pairs of attributes
    // `text` element text
    node(name, attributes, text) {
      var child;
      if (name != null) {
        name = name.valueOf();
      }
      attributes || (attributes = {});
      attributes = attributes.valueOf();
      // swap argument order: text <-> attributes
      if (!isObject(attributes)) {
        [text, attributes] = [attributes, text];
      }
      child = new XMLElement(this, name, attributes);
      if (text != null) {
        child.text(text);
      }
      this.children.push(child);
      return child;
    }

    // Creates a text node

    // `value` element text
    text(value) {
      var child;
      child = new XMLText(this, value);
      this.children.push(child);
      return this;
    }

    // Creates a CDATA node

    // `value` element text without CDATA delimiters
    cdata(value) {
      var child;
      child = new XMLCData(this, value);
      this.children.push(child);
      return this;
    }

    // Creates a comment node

    // `value` comment text
    comment(value) {
      var child;
      child = new XMLComment(this, value);
      this.children.push(child);
      return this;
    }

    // Creates a comment node before the current node

    // `value` comment text
    commentBefore(value) {
      var child, i, removed;
      // temporarily remove children starting *with* this
      i = this.parent.children.indexOf(this);
      removed = this.parent.children.splice(i);
      // add the new child
      child = this.parent.comment(value);
      // add back removed children after new child
      Array.prototype.push.apply(this.parent.children, removed);
      return this;
    }

    // Creates a comment node after the current node

    // `value` comment text
    commentAfter(value) {
      var child, i, removed;
      // temporarily remove children starting *after* this
      i = this.parent.children.indexOf(this);
      removed = this.parent.children.splice(i + 1);
      // add the new child
      child = this.parent.comment(value);
      // add back removed children after new child
      Array.prototype.push.apply(this.parent.children, removed);
      return this;
    }

    // Adds unescaped raw text

    // `value` text
    raw(value) {
      var child;
      child = new XMLRaw(this, value);
      this.children.push(child);
      return this;
    }

    // Adds a processing instruction

    // `target` instruction target
    // `value` instruction value
    instruction(target, value) {
      var insTarget, insValue, instruction, j, len;
      if (target != null) {
        target = target.valueOf();
      }
      if (value != null) {
        value = value.valueOf();
      }
      if (Array.isArray(target)) { // expand if array
        for (j = 0, len = target.length; j < len; j++) {
          insTarget = target[j];
          this.instruction(insTarget);
        }
      } else if (isObject(target)) { // expand if object
        for (insTarget in target) {
          if (!hasProp.call(target, insTarget)) continue;
          insValue = target[insTarget];
          this.instruction(insTarget, insValue);
        }
      } else {
        if (isFunction(value)) {
          value = value.apply();
        }
        instruction = new XMLProcessingInstruction(this, target, value);
        this.children.push(instruction);
      }
      return this;
    }

    // Creates a processing instruction node before the current node

    // `target` instruction target
    // `value` instruction value
    instructionBefore(target, value) {
      var child, i, removed;
      // temporarily remove children starting *with* this
      i = this.parent.children.indexOf(this);
      removed = this.parent.children.splice(i);
      // add the new child
      child = this.parent.instruction(target, value);
      // add back removed children after new child
      Array.prototype.push.apply(this.parent.children, removed);
      return this;
    }

    // Creates a processing instruction node after the current node

    // `target` instruction target
    // `value` instruction value
    instructionAfter(target, value) {
      var child, i, removed;
      // temporarily remove children starting *after* this
      i = this.parent.children.indexOf(this);
      removed = this.parent.children.splice(i + 1);
      // add the new child
      child = this.parent.instruction(target, value);
      // add back removed children after new child
      Array.prototype.push.apply(this.parent.children, removed);
      return this;
    }

    // Creates the xml declaration

    // `version` A version number string, e.g. 1.0
    // `encoding` Encoding declaration, e.g. UTF-8
    // `standalone` standalone document declaration: true or false
    declaration(version, encoding, standalone) {
      var doc, xmldec;
      doc = this.document();
      xmldec = new XMLDeclaration(doc, version, encoding, standalone);
      // Replace XML declaration if exists, otherwise insert at top
      if (doc.children[0] instanceof XMLDeclaration) {
        doc.children[0] = xmldec;
      } else {
        doc.children.unshift(xmldec);
      }
      return doc.root() || doc;
    }

    // Creates the document type declaration

    // `pubID` the public identifier of the external subset
    // `sysID` the system identifier of the external subset
    doctype(pubID, sysID) {
      var child, doc, doctype, i, j, k, len, len1, ref, ref1;
      doc = this.document();
      doctype = new XMLDocType(doc, pubID, sysID);
      ref = doc.children;
      // Replace DTD if exists
      for (i = j = 0, len = ref.length; j < len; i = ++j) {
        child = ref[i];
        if (child instanceof XMLDocType) {
          doc.children[i] = doctype;
          return doctype;
        }
      }
      ref1 = doc.children;
      // insert before root node if the root node exists
      for (i = k = 0, len1 = ref1.length; k < len1; i = ++k) {
        child = ref1[i];
        if (child.isRoot) {
          doc.children.splice(i, 0, doctype);
          return doctype;
        }
      }
      // otherwise append to end
      doc.children.push(doctype);
      return doctype;
    }

    // Gets the parent node
    up() {
      if (this.isRoot) {
        throw new Error("The root node has no parent. Use doc() if you need to get the document object.");
      }
      return this.parent;
    }

    // Gets the root node
    root() {
      var node;
      node = this;
      while (node) {
        if (node.isDocument) {
          return node.rootObject;
        } else if (node.isRoot) {
          return node;
        } else {
          node = node.parent;
        }
      }
    }

    // Gets the node representing the XML document
    document() {
      var node;
      node = this;
      while (node) {
        if (node.isDocument) {
          return node;
        } else {
          node = node.parent;
        }
      }
    }

    // Ends the document and converts string
    end(options) {
      return this.document().end(options);
    }

    // Gets the previous node
    prev() {
      var i;
      i = this.parent.children.indexOf(this);
      if (i < 1) {
        throw new Error("Already at the first node");
      }
      return this.parent.children[i - 1];
    }

    // Gets the next node
    next() {
      var i;
      i = this.parent.children.indexOf(this);
      if (i === -1 || i === this.parent.children.length - 1) {
        throw new Error("Already at the last node");
      }
      return this.parent.children[i + 1];
    }

    // Imports cloned root from another XML document

    // `doc` the XML document to insert nodes from
    importDocument(doc) {
      var clonedRoot;
      clonedRoot = doc.root().clone();
      clonedRoot.parent = this;
      clonedRoot.isRoot = false;
      this.children.push(clonedRoot);
      return this;
    }

    // Aliases
    ele(name, attributes, text) {
      return this.element(name, attributes, text);
    }

    nod(name, attributes, text) {
      return this.node(name, attributes, text);
    }

    txt(value) {
      return this.text(value);
    }

    dat(value) {
      return this.cdata(value);
    }

    com(value) {
      return this.comment(value);
    }

    ins(target, value) {
      return this.instruction(target, value);
    }

    doc() {
      return this.document();
    }

    dec(version, encoding, standalone) {
      return this.declaration(version, encoding, standalone);
    }

    dtd(pubID, sysID) {
      return this.doctype(pubID, sysID);
    }

    e(name, attributes, text) {
      return this.element(name, attributes, text);
    }

    n(name, attributes, text) {
      return this.node(name, attributes, text);
    }

    t(value) {
      return this.text(value);
    }

    d(value) {
      return this.cdata(value);
    }

    c(value) {
      return this.comment(value);
    }

    r(value) {
      return this.raw(value);
    }

    i(target, value) {
      return this.instruction(target, value);
    }

    u() {
      return this.up();
    }

    // can be deprecated in a future release
    importXMLBuilder(doc) {
      return this.importDocument(doc);
    }

  };

}).call(this);
